from clarify_assets.broker.broker import AssetBroker
from clarify_assets.models.asset import Asset
from clarify_assets.models.attribute import Attribute
from pyclarify import Item

__version__ = "0.0.7"



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Optional

from ..__utils__.copy import jsoncopy, merge_filters
from ..__utils__.parsers import parse_to_clarify_filter

from pyclarify import Item, Signal, query


@dataclass
class Attribute:
    meta: Item
    filter: query.Filter = field(default_factory=dict)
    add_filter: query.Filter = field(default_factory=dict)
    signals: Optional[list[Signal]] = field(default_factory=list)
    independent: bool = False  # Ignore parent filters
    name: str = None
    labels: list[str] = field(default_factory=list)

    def __post_init__(self):
        if self.name:
            if self.meta:
                self.meta.name = self.name
            else:
                self.meta = Item(name=self.name)
        if self.add_filter:
            self.filter = merge_filters(self.filter, self.add_filter)
            self.add_filter = {}

    def copy(
        self,
        meta=None,
        name=None,
        labels=None,
        filter=None,
        signals=None,
        independent=None,
    ):
        if not meta:
            meta = deepcopy(self.meta)
        if name:
            meta.name = name
        if labels:
            meta.labels = labels
        if not filter:
            filter = jsoncopy(self.filter)
        if not signals:
            signals = deepcopy(self.signals)
        if not independent:
            independent = self.independent
        return Attribute(meta, filter, signals, independent=independent)

    def to_clarify_filter(self):
        return parse_to_clarify_filter(self.filter)

    def add_signal(self, signals):
        if signals:
            if not isinstance(signals, list):
                signals = [signals]
            else:
                if isinstance(signals[0], list):
                    signals = [s for l in signals for s in l]
            for signal in signals:
                if signal not in self.signals:
                    self.signals += [signal]

    def __str__(self):
        return self.meta.name

    def __hash__(self):
        return hash(tuple(self))



from pyclarify.fields.query import Operators

def can_be_published(leaf):
    return leaf.signals != []


def get_item_id_or_none(leaf):
    if leaf.signals:
        if leaf.signals[0].relationships:
            if leaf.signals[0].relationships.item:
                if leaf.signals[0].relationships.item.data:
                    if leaf.signals[0].relationships.item.data.id:
                        return leaf.signals[0].relationships.item.data.id
    return None


def compare_labels(leaf, item):
    item_labels = []
    leaf_labels = []
    for k, v in leaf.meta.labels.items():
        if v:
            leaf_labels.append((k, v[0]))
        else:
            leaf_labels.append((k, ""))
    for k, v in item.attributes.labels.items():
        if v:
            item_labels.append((k, v[0]))
        else:
            item_labels.append((k, ""))
    set1 = set(leaf_labels)
    set2 = set(item_labels)
    new_label = list(set1 - set2)
    removed_label = list(set2 - set1)
    return new_label, removed_label


def compare_meta_and_published_item(leaf, item):
    new_labels, removed_labels = compare_labels(leaf, item)
    if leaf.meta.name != item.attributes.name:
        new_labels.append(("name", leaf.meta.name))
        removed_labels.append(("name", item.attributes.name))
    if leaf.meta.engUnit != item.attributes.engUnit:
        new_labels.append(("engUnit", leaf.meta.engUnit))
        removed_labels.append(("engUnit", item.attributes.engUnit))
    if leaf.meta.valueType != item.attributes.valueType:
        new_labels.append(("valueType", leaf.meta.valueType))
        removed_labels.append(("valueType", item.attributes.valueType))
    return new_labels, removed_labels


def is_leaf(obj):
    if isinstance(obj, dict):
        # check value
        keys = list(obj.keys())
        if len(keys) != 1:
            return False

        # check only child
        value = obj[keys[0]]
        if keys[0].startswith("labels"):
            return True

        # is string obj is eq
        if isinstance(value, str):
            return True

        if isinstance(value, list):
            return False

        if isinstance(value, dict):
            # check for operator
            keys = list(value.keys())

            if len(keys) != 1:
                return False

            try:
                op = Operators(keys[0])
                return True
            except:
                return False



from dataclasses import dataclass, field
from typing import Optional
from pyclarify import query
from copy import deepcopy

from ..__utils__.copy import jsoncopy, merge_filters, merge_labels
from ..__utils__.parsers import parse_to_clarify_filter
from .attribute import Attribute


@dataclass
class Asset:
    parent: Optional["Asset"] = None  # Can only have one parent
    children: Optional[list["Asset"]] = field(
        default_factory=list
    )  # can have multiple children
    labels: list[str] = field(default_factory=list)
    filter: list[query.Filter] = field(default_factory=list)
    attributes: Optional[list[Attribute]] = field(default_factory=list)
    independent: bool = False

    def __post_init__(self):
        if self.parent:
            self.labels = merge_labels(self.labels, self.parent.labels)
            if not self.independent:
                self.filter = merge_filters(self.filter, self.parent.filter)

            if self.parent.children:
                if self not in self.parent.children:
                    self.parent.children += [self]
            else:
                self.parent.children = [self]

            # EXPERIMENTAL
            attr_name = list(self.labels[0].values())[0].lower().replace(" ", "_")
            if not hasattr(self.parent, attr_name):
                setattr(self.parent, attr_name, self)

        # update filter for attributes
        if self.attributes:
            for attr in self.attributes:
                attr_name = str(attr)
                if not hasattr(self.parent, attr_name):
                    setattr(self, attr_name, attr)
                attr.meta.labels = merge_labels(
                    attr.labels, self.labels, return_dictionary=True
                )
                attr.labels = merge_labels(attr.labels, self.labels)

                if not attr.independent:  # ignore parent filters
                    if attr.filter:
                        attr.filter = merge_filters(attr.filter, self.filter)
                    else:
                        attr.filter = self.filter

    def __eq__(self, other):
        return self.labels == other.labels and self.filter == other.filter

    def __hash__(self):
        return hash(str(self))

    def to_clarify_filter(self):
        return parse_to_clarify_filter(self.filter)

    def append(self, child_asset_or_attribute):
        if isinstance(child_asset_or_attribute, list):
            for obj in child_asset_or_attribute:
                self.append(obj)
        elif isinstance(child_asset_or_attribute, self.__class__):
            self.children.append(child_asset_or_attribute)
            self.link_tree()
            self.children = list(set(self.children))
        elif isinstance(child_asset_or_attribute, Attribute):
            self.attributes.append(child_asset_or_attribute)
            self.link_tree()
            self.attributes = list(set(self.attributes))

    def get_root(self):
        if self.parent:
            return self.parent.get_root()
        else:
            return self

    def assert_parents(self):
        if self.children:
            for child in self.children:
                child.parent = self
                self.__post_init__()
                child.assert_parents()
        else:
            self.__post_init__()
            if self.attributes:
                for attr in self.attributes:
                    attr.__post_init__()

    def link_tree(self):
        root = self.get_root()
        root.assert_parents()

    def __str__(self):
        return str(list(self.labels[0].values())[0])

    def generate_string(
        self, debug=False, minimal=False, only_failures=False, indent=0
    ):
        STAR = " \u2605 "
        CHECK = " \u2713 "
        CROSS = " \u2718 "
        ARROW = " \u2192 "
        GREEN = "\033[1;32m"
        ORANGE = "\033[1;33m"
        RED = "\033[1;31m"
        NEW_LINE_RESET = "\033[0;30m\n"
        string = ""
        current_asset = list(self.labels[0].values())[0]

        if self.attributes:
            if minimal:
                if len(self.attributes) == sum(
                    [len(attr.signals) for attr in self.attributes]
                ):
                    if not only_failures:
                        string += GREEN + current_asset + CHECK

                else:
                    string += RED + current_asset + CROSS
            if debug:
                if not minimal:
                    string += current_asset + NEW_LINE_RESET
                    indent += 1
                    # TODO: Add sorting of attributes
                    for attr in self.attributes:
                        string += ARROW * indent + str(attr.meta.name) + "\n"
                        if attr.signals:
                            if len(attr.signals) == 1:
                                string += (
                                    ARROW * (indent + 1)
                                    + STAR
                                    + GREEN
                                    + str(attr.signals[0].attributes.name)
                                    + NEW_LINE_RESET
                                )
                            else:
                                string += (
                                    ARROW * (indent + 1)
                                    + STAR
                                    + RED
                                    + str(attr.to_clarify_filter())
                                    + NEW_LINE_RESET
                                )
                                if debug:
                                    for s in attr.signals:
                                        string += (
                                            ARROW * (indent + 1)
                                            + STAR
                                            + RED
                                            + str(s.attributes.name)
                                            + str(s.attributes.labels)
                                            + NEW_LINE_RESET
                                        )
                        else:
                            # Print filter as well
                            string += (
                                ARROW * (indent + 1)
                                + STAR
                                + ORANGE
                                + str(attr.to_clarify_filter())
                                + NEW_LINE_RESET
                            )
                    indent -= 1
                else:
                    if not len(self.attributes) == sum(
                        [len(attr.signals) for attr in self.attributes]
                    ):
                        string += NEW_LINE_RESET
                        indent += 1
                        # TODO: Add sorting of attributes
                        for attr in self.attributes:
                            string += ARROW * indent + str(attr.meta.name) + "\n"
                            if attr.signals:
                                if len(attr.signals) == 1:
                                    string += (
                                        ARROW * (indent + 1)
                                        + STAR
                                        + GREEN
                                        + str(attr.signals[0].attributes.name)
                                        + NEW_LINE_RESET
                                    )
                                else:
                                    string += (
                                        ARROW * (indent + 1)
                                        + STAR
                                        + RED
                                        + str(attr.to_clarify_filter())
                                        + NEW_LINE_RESET
                                    )
                                    if debug:
                                        for s in attr.signals:
                                            string += (
                                                ARROW * (indent + 1)
                                                + STAR
                                                + RED
                                                + str(s.attributes.name)
                                                + str(s.attributes.labels)
                                                + NEW_LINE_RESET
                                            )
                            else:
                                # Print filter as well
                                string += (
                                    ARROW * (indent + 1)
                                    + STAR
                                    + ORANGE
                                    + str(attr.to_clarify_filter())
                                    + NEW_LINE_RESET
                                )
                        indent -= 1
        else:
            string += current_asset
        if self.children:
            indent += 1
            # TODO: Add sorting of children
            if string == "":
                string += current_asset
            for child in self.children:
                string += child.generate_string(
                    debug=debug,
                    minimal=minimal,
                    only_failures=only_failures,
                    indent=indent,
                )
        indent -= 1
        if string != "" and string != NEW_LINE_RESET:
            string = NEW_LINE_RESET + ARROW * indent + string + NEW_LINE_RESET
        return string

    def pretty_print(self, debug=False, minimal=False, only_failures=False):
        # Clean up some logic
        if only_failures:
            debug = True
            minimal = True

        print(
            self.generate_string(
                debug=debug, minimal=minimal, only_failures=only_failures, indent=0
            )
        )

    def copy(
        self, parent=None, children=None, labels=None, filter=None, attributes=None
    ):
        if not parent:
            parent = deepcopy(self.parent)
        if not children:
            children = deepcopy(self.children)
        if not labels:
            labels = jsoncopy(self.labels)
        if not filter:
            filter = jsoncopy(self.filter)
        if not attributes:
            attributes = deepcopy(self.attributes)
        return Asset(parent, children, labels, filter, attributes)



import json


def jsoncopy(dictionary):
    return json.loads(json.dumps(dictionary))


def merge_labels(self_labels, other_labels, return_dictionary=False):
    # returns list by default
    labels = []
    if self_labels:
        labels += self_labels
    if other_labels:
        labels += other_labels
    dictionary = {}
    for label in labels:
        label = {list(label.keys())[0]: [list(label.values())[0]]}
        dictionary = dictionary | label
    if return_dictionary:
        return dictionary
    return [{k: v[0]} for k, v in dictionary.items()]


def merge_filters(self_filter, other_filter):
    """
    :meta private:

    Merges two clarify filters to a single filter

    Parameters
    ----------
        self_filter: Dict
            ...

        other_filter: Dict
            ...
    Example
    -------
        ...

    """
    self_filter = jsoncopy(self_filter)
    parent_filter = jsoncopy(other_filter)
    if self_filter and parent_filter:
        if "$and" in self_filter.keys():
            self_filter = self_filter["$and"]
        else:
            self_filter = [self_filter]

        if "$and" in parent_filter.keys():
            parent_filter = parent_filter["$and"]
        else:
            parent_filter = [parent_filter]

        self_keys = [list(sf.keys())[0] for sf in self_filter]

        for f in parent_filter:
            f_key = list(f.keys())[0]
            if f_key not in self_keys:
                self_filter += [f]
            else:
                indx = self_keys.index(f_key)
                # pop overlapping element
                self_filter.pop(indx)
                self_filter.insert(indx, f)
        if len(self_filter) == 1:
            return self_filter[0]
        return {"$and": self_filter}

    elif parent_filter:
        return parent_filter
    return self_filter



from .copy import jsoncopy
from .validators import is_leaf
import json


def parse_to_sqlite3(obj, key_list, hash_key=None):
    format = []
    for k in key_list:
        if k == "HashKey":
            v = hash_key
        elif k in obj.keys():
            v = obj[k]
        else:
            v = None
        if isinstance(v, list):
            if v != []:
                v = v[0]
            else:
                v = None
        format.append(v)
    return format


# Modifying signals for lookups
def flatten(signal):
    d_signal = signal.model_dump(mode="json")
    d_signal = d_signal.pop("attributes")
    d_signal["id"] = signal.id
    # DO NOT SUPPORT ENUM VALUES
    d_signal.pop("enumValues")
    label_copy = jsoncopy(d_signal["labels"])

    for key in label_copy.keys():
        d_signal[key] = label_copy[key]
    del label_copy
    d_signal.pop("labels")
    return d_signal


def scuffed_filter_key_extraction(filter):
    if "$and" in filter.keys():
        filters = str(filter["$and"]).split(",")
    else:
        filters = list(filter.keys())

    filter_keys = []
    for f in filters:
        f = f.strip()

        if f[0] == "[":
            f = f[1:]
        if f[-1] == "]":
            f = f[:-1]
        try:
            f = eval(f)
            f = list(f.keys())[0]
            f = f.replace("labels.", "").replace("-", "_") # cant use dash in SQL keys
            filter_keys += [f]
        except:
            pass
    return filter_keys


def parse_filter(filter, table):
    key = list(filter.keys())[0]
    value = list(filter.values())[0]
    if key == "$and":
        filter = filter["$and"]
        string = ""
        for _filter in filter:
            string += parse_filter(_filter, table) + " AND "
        return string[:-5]
    elif key == "$or":
        filter = filter["$or"]
        for _filter in filter:
            string += parse_filter(_filter, table) + " OR "
        return string[:-4]
    else:

        operator = "$eq"

        if isinstance(value, dict):
            operator = list(value.keys())[0]
            value = list(value.values())[0]
        key = key.replace("-", "_").replace("labels.", "")
        match operator:
            case "$eq":
                if isinstance(value, list):
                    value = value[0]
                return f"{key} == '{value}'"

            case "$ne":
                if isinstance(value, list):
                    value = value[0]
                return f"{key} != '{value}'"

            case "$in":
                return f'{key} IN {str(value).replace("[", "(").replace("]", ")")}'

            case "$nin":
                return f'id NOT IN (SELECT id FROM {table} where {key} IN {str(value).replace("[", "(").replace("]", ")")})'

            case "$regex":
                if isinstance(value, list):
                    value = value[0]
                return f"REGEXP({key}, '{value}')"
        return False


def get_all_keys(look_up_list):
    set_of_keys = set()
    for item in look_up_list:
        for k in item.keys():
            set_of_keys.add(k)
    set_of_keys.add("HashKey")
    return set_of_keys


def update_leaf_value(leaf):
    key = list(leaf.keys())[0]
    value = list(leaf.values())[0]
    if isinstance(value, list):
        value = value[0]
    return value


def iterate_sub_dict(s):
    key = list(s.keys())[0]
    value = list(s.values())[0]
    if is_leaf(s):
        # clean_labels
        if key.startswith("labels"):
            value = update_leaf_value(s)
        return {key: value}
    else:
        return {key: [iterate_sub_dict(s) for s in value]}


def parse_to_clarify_filter(filter):
    my_new_dict = {}
    if filter:
        tmp_filter = jsoncopy(filter)
        my_new_dict = iterate_sub_dict(tmp_filter)
    return json.dumps(my_new_dict)



import re
import tempfile
from time import sleep
from tqdm import tqdm

from pyclarify import Client
from pyclarify.experimental import ExperimentalClient
from pyclarify.fields.resource import (
    RelationshipsDictSignal,
    RelationshipDataToOne,
    Identifier,
)

from ..__utils__.parsers import (
    parse_filter,
    flatten,
    get_all_keys,
    parse_to_sqlite3,
    scuffed_filter_key_extraction,
)
from ..__utils__.validators import (
    get_item_id_or_none,
    compare_meta_and_published_item,
    can_be_published,
)
from ..__utils__.printing import color


def iterate_nodes(asset):
    leaves = []
    if asset.children:
        for child in asset.children:
            leaves += iterate_nodes(child)
    if asset.attributes:
        for attribute in asset.attributes:
            leaves += [attribute]
    return leaves


def make_query(filter, table="signals"):
    q = f"SELECT HashKey from {table} WHERE " + parse_filter(filter, table)
    return q


class AssetBroker:
    def __init__(self, credentials, integrations):
        self.credentials = credentials
        self.INTEGRATIONS = integrations
        self.client = Client(credentials)
        self.exp_client = ExperimentalClient(credentials)
        self.SIGNALS = []
        self.ITEMS = []
        self.look_up_signals = None
        self.look_up_items = None
        self.SET_OF_SIGNAL_KEYS = None
        self.SET_OF_ITEM_KEYS = None
        self.con = None
        self.cur = None
        self.hash_map_signals = None
        self.hash_map_items = None

        self.tmp_folder = tempfile.TemporaryDirectory()
        self.tmp_folder_path = self.tmp_folder.name

        # Used for item comparisons
        self.MATCH_LABELS=["site", "area", "line", "cell", "data-source"]

    def get_data_from_clarify(self, reset=False):
        if not reset:
            if self.SIGNALS and self.ITEMS:
                return
        
        else:
            self.SIGNALS = []
            self.ITEMS = []
        signals = []
        for integration in self.INTEGRATIONS:
            r = self.client.select_signals(
                integration=integration, limit=99999, include=["item"]
            )
            if r.error:
                num_errors = len(r.error)
            else:
                num_errors = 0
            if r.result:
                if r.result.data:
                    num_signals = len(r.result.data)
                    for s in r.result.data:
                        s.relationships.integration = RelationshipsDictSignal(
                            integration=RelationshipDataToOne(
                                data=Identifier(type="integration", id=integration)
                            )
                        )
                        signals += [s]
                else:
                    num_signals = 0
            else:
                print(r.error)
            print(
                f"Integration: {integration} - signals: {num_signals} - errors: {num_errors}"
            )
            sleep(0.5)

        self.SIGNALS = signals

        items = []

        r = self.client.select_items(limit=99999)
        items = r.result.data
        print(f"Retrieved {len(items)} items")

        self.ITEMS += items

        # POST PROCESS
        self.look_up_signals = [flatten(s) for s in self.SIGNALS]
        self.look_up_items = [flatten(i) for i in self.ITEMS]
        self.SET_OF_SIGNAL_KEYS = get_all_keys(self.look_up_signals)
        self.SET_OF_ITEM_KEYS = get_all_keys(self.look_up_items)

    def set_up_db(self):
        # check if already has db
        if self.con:
            import tempfile

            self.cur.close()
            self.con.close()
            self.tmp_folder.cleanup()
            self.tmp_folder = tempfile.TemporaryDirectory()
            self.tmp_folder_path = self.tmp_folder.name

        import sqlite3
        import uuid
        import hashlib

        con = sqlite3.connect(f"{self.tmp_folder_path}/{uuid.uuid4()}.db")
        print("Created SQLite3 db")
        cur = con.cursor()
        cur.execute("DROP TABLE IF EXISTS items;")
        cur.execute("DROP TABLE IF EXISTS signals;")
        # ADD REGEX FUNCTIONALITY

        def functionRegex(value, pattern):
            c_pattern = re.compile(r"\b" + pattern + r"\b")
            if value is not None:
                return c_pattern.search(value) is not None
            return False

        con.create_function("REGEXP", 2, functionRegex)

        cur.execute(
            f"CREATE TABLE items({', '.join(self.SET_OF_ITEM_KEYS).replace('-', '_')})"
        )
        cur.execute(
            f"CREATE TABLE signals({', '.join(self.SET_OF_SIGNAL_KEYS).replace('-', '_')})"
        )

        data = []
        hash_map_signals = {}
        for lus, s in zip(self.look_up_signals, self.SIGNALS):
            hash_key = hashlib.sha256(str(lus).encode("utf-8")).hexdigest()

            hash_map_signals[hash_key] = s
            data.append(parse_to_sqlite3(lus, self.SET_OF_SIGNAL_KEYS, hash_key))

        inputs = "?, " * len(self.SET_OF_SIGNAL_KEYS)
        inputs = inputs[:-2]
        cur.executemany(f"INSERT INTO signals VALUES ({inputs})", data)

        data = []
        hash_map_items = {}
        for lui, i in zip(self.look_up_items, self.ITEMS):
            hash_key = hashlib.sha256(str(lui).encode("utf-8")).hexdigest()

            hash_map_items[hash_key] = i
            data.append(parse_to_sqlite3(lui, self.SET_OF_ITEM_KEYS, hash_key))

        inputs = "?, " * len(self.SET_OF_ITEM_KEYS)
        inputs = inputs[:-2]
        cur.executemany(f"INSERT INTO items VALUES ({inputs})", data)

        con.commit()
        print("Finished inserting data")
        self.con = con
        self.cur = cur
        self.hash_map_signals = hash_map_signals
        self.hash_map_items = hash_map_items

    def flush_and_build_db(self, reset_clarify=False):
        self.get_data_from_clarify(reset=reset_clarify)
        self.set_up_db()

    def is_valid_filter(self, filter, table):
        # Very Scuffed not happy
        valid_keys = getattr(self, f"SET_OF_{table.upper()[0:-1]}_KEYS")
        for k in scuffed_filter_key_extraction(filter):
            if k not in valid_keys:
                return False
        return True

    def query_sqlite(self, filter, table="signals"):
        if not self.is_valid_filter(filter, table):
            return []
        q = make_query(filter, table=table)

        res = self.cur.execute(q)
        return res.fetchall()

    def assign_signals(self, asset, force_reset=False):
        leaves = iterate_nodes(asset)
        if force_reset:
            for leaf in leaves:
                leaf.signals = []

        for leaf in tqdm(leaves, leave=True):
            if not force_reset:
                if leaf.signals:
                    continue
            hash_keys_ids = self.query_sqlite(leaf.filter)
            signals = []
            for hash_key_id in hash_keys_ids:
                hash_key_id = hash_key_id[0]
                signals.append(self.hash_map_signals[hash_key_id])
            if signals:
                leaf.add_signal(signals)

    def check_for_published_item(
        self, leaf
    ):
        item = leaf.meta
        jattr = item.model_dump(mode="json")

        # Check necessary fields
        # Name, Area, maybe Line, maybe Cell
        and_list = [{"name": jattr["name"]}]
        for label in self.MATCH_LABELS:
            if label in jattr["labels"].keys():
                and_list += [{f"labels.{label}": jattr["labels"][label]}]
        filt = {"$and": and_list}
        hash_keys_ids = self.query_sqlite(filt, "items")
        items = []
        for hash_key_id in hash_keys_ids:
            hash_key_id = hash_key_id[0]
            possible_items = self.hash_map_items[hash_key_id]
            if not isinstance(possible_items, list):
                items.append(possible_items)

        # Check self.signals
        possible_item_id = get_item_id_or_none(leaf)
        if possible_item_id:
            possible_item = list(
                filter(lambda i: possible_item_id == i.id, self.ITEMS)
            )[0]

            if possible_item not in items:
                items.append(possible_item)

        items = sorted(items, key=lambda i: i.meta.createdAt, reverse=True)

        return items

    def publish_leaves(self, leaves):
        for integration in self.INTEGRATIONS:
            sub_leaves = list(
                filter(
                    lambda x: (
                        x.signals[0].relationships.integration.integration.data.id
                        == integration
                        if x.signals
                        else False
                    ),
                    leaves,
                )
            )
            sub_leaves = {l.signals[0].id: l.meta for l in sub_leaves}
            r = self.client.publish_signals(
                items_by_signal=sub_leaves, integration=integration
            )
            print(r)

    def connect_signal_to_item(self, signal, item):
        source_signal_id = signal.id
        source_integration = signal.relationships.integration.integration.data.id

        print(f"connecting {source_signal_id} to {item.id}")
        r = self.exp_client.connect_signals(
            filter={
                "id": source_signal_id,
            },
            integration=source_integration,
            item=item.id,
            # include=["item"],
            dryrun=False,
        )

    def connect_leaves_to_items(self, old_leaves):
        for leaf, item in old_leaves:
            possible_item_id = get_item_id_or_none(leaf)
            if possible_item_id:
                if possible_item_id != item.id:
                    self.connect_signal_to_item(leaf.signals[0], item)
            else:
                self.connect_signal_to_item(leaf.signals[0], item)

    def disconnect_signal(self, signal):
        source_signal_id = signal.id
        source_integration = signal.relationships.integration.integration

        # disconnect from old item
        r = self.exp_client.disconnect_signals(
            filter={
                "id": source_signal_id,
            },
            integration=source_integration,
            dryrun=False,
        )

    def disconnect_and_reconnect(self, signal, new_item, old_item):
        source_signal = list(
            filter(
                lambda s: (
                    old_item.id in s.relationships.item.data.id
                    if s.relationships.item.data
                    else False
                ),
                self.SIGNALS,
            )
        )[0]
        self.disconnect_signal(source_signal)
        self.connect_signal_to_item(source_signal, new_item)

    def publish_asset(self, asset):
        leaves = iterate_nodes(asset)
        new_leaves = []
        old_leaves = []
        new_meta = []  # For future
        conflicting_leaves = []

        for leaf in leaves:
            if not can_be_published(leaf):
                continue

            # Check for existing items
            items = self.check_for_published_item(leaf)

            if items:
                if len(items) == 1:
                    possible_id = get_item_id_or_none(leaf)
                    if items[0].id == possible_id:
                        new_labels, removed_labels = compare_meta_and_published_item(
                            leaf, items[0]
                        )
                        if new_labels or removed_labels:
                            old_leaves.append((leaf, items[0]))
                    else:
                        old_leaves.append((leaf, items[0]))
                else:
                    conflicting_leaves.append((leaf, items))
            else:
                new_leaves.append(leaf)
        if new_leaves:
            print("New Items")
            print(f"{color.BOLD}{'Signal Name':30}{'Labels':60}{color.END}")
            for leaf in new_leaves:
                labels = list(map(lambda l: list(l.values())[0], leaf.labels))[::-1]
                print(f"{str(leaf):30}{' -> '.join(labels)}")

            ans = input("\nPublish signals? y/n \n")
            if ans == "y":
                print("Publishing all new signals")
                self.publish_leaves(new_leaves)
            else:
                print("You declined")

        print()
        if old_leaves:
            print("Signals with existing Item")
            print(
                f"{color.BOLD}{'Signal Name':30}{'Existing Id':30}{'Labels':40}{'Removed labels':50}{'New labels'}{color.END}"
            )
            for leaf, item in old_leaves:
                new_labels, removed_labels = compare_meta_and_published_item(leaf, item)

                new_labels_formatted = " | ".join(
                    [
                        f'{nl[0]}:{nl[-1] if nl[-1] else f"{color.RED}Null{color.END}"}'
                        for nl in new_labels
                    ]
                )
                removed_labels_formatted = " | ".join(
                    [f'{rl[0]}:{rl[-1] if rl[-1] else "Null"}' for rl in removed_labels]
                )

                labels = list(map(lambda l: list(l.values())[0], leaf.labels))[:2:][
                    ::-1
                ]
                curr_labels_formatted = " -> ".join(labels)
                print(
                    f"{str(leaf):30}{item.id:30}{curr_labels_formatted:40}{removed_labels_formatted:50}{new_labels_formatted}"
                )

            ans = input("\nConnect signal sources? y/n\n")
            if ans == "y":
                self.connect_leaves_to_items(old_leaves)
                ans = input("\nMerge(m) or overwrite(o) conflicting metas? m/o\n")
                if ans == "m":
                    ans = input("\nOld has precedence? y/n\n")
                    if ans == "y":
                        old_precedes = True
                    else:
                        old_precedes = False

                    for leaf, item in old_leaves:
                        new_labels, removed_labels = compare_meta_and_published_item(
                            leaf, item
                        )
                        preceded_label = new_labels if old_precedes else removed_labels
                        preceding_label = removed_labels if old_precedes else new_labels
                        updated_dict = dict(set(preceded_label) ^ set(preceding_label))
                        for k, v in updated_dict.items():
                            if k == "name":
                                leaf.meta.name = v
                            elif k == "engUnit":
                                leaf.meta.engUnit = v
                            else:
                                leaf.meta.labels[k] = [v]

                    self.publish_leaves([leaf for leaf, item in old_leaves])

                elif ans == "o":
                    self.publish_leaves([leaf for leaf, item in old_leaves])
            else:
                print("You declined")

        print()
        if conflicting_leaves:
            print("Conflicting Signals")
            print(
                f"{color.BOLD}{'Signal Name':30}{'Existing Id':30}{'Labels':40}{'Removed labels':50}{'New labels'}{color.END}"
            )
            for leaf, items in conflicting_leaves:
                for item in items:
                    new_labels, removed_labels = compare_meta_and_published_item(
                        leaf, item
                    )

                    new_labels_formatted = " | ".join(
                        [
                            f'{nl[0]}:{nl[-1] if nl[-1] else f"{color.RED}Null{color.END}"}'
                            for nl in new_labels
                        ]
                    )
                    removed_labels_formatted = " | ".join(
                        [
                            f'{rl[0]}:{rl[-1] if rl[-1] else "Null"}'
                            for rl in removed_labels
                        ]
                    )

                    labels = list(map(lambda l: list(l.values())[0], leaf.labels))[:2:][
                        ::-1
                    ]
                    curr_labels_formatted = " -> ".join(labels)
                    print(
                        f"{str(leaf):30}{item.id:30}{curr_labels_formatted:40}{removed_labels_formatted:50}{new_labels_formatted}"
                    )

            ans = input(
                "Disconnect signal from old item and connect to new item? y/n \n"
            )

            if ans == "y" and False:  # Skipping this for now
                print("Disconnecting and reconnecting signals")
                self.disconnect_and_reconnect(leaf.signals[0], item[0], item[1])
                print("Remember to delete/update old items!")
            else:
                print("you declined")
